C
Code by Michael Kupferschmid
C
      SUBROUTINE BISECT(FCN,TOL, XL,XR,NFE, X,F,RC)
C     This routine finds the value of X between XL and XR
C     where the function FCN is zero, and returns in F the
C     value of the function at that point.
C
C     RC < 0 on input suppresses error messages.
C     RC output  meaning
C     ---------  -------
C     0          convergence in both X and F
C     1          convergence in X or F is the best we can do
C     2          no root in the interval
C     3          failure to converge in NFE function evaluations
C     4          failure to converge in 100 bisections
C
C     variable  meaning
C     --------  -------
C     DABS      Fortran function returns |REAL*8|
C     DMAX1     Fortran function returns larger of REAL*8s
C     F         function value at current trial point X
C     FCN       routine computes value of function
C     FL        function value at XL
C     FR        function value at XR
C     FTOL      F tolerance used
C     I         index on the bisections
C     NF        number of function evaluations used
C     NFE       number of function evaluations allowed
C     RC        return code; see table above
C     TOL       tolerances on X and F
C     X         midpoint of XL and XR
C     XL        left endpoint of interval containing root
C     XR        right endpoint of interval containing root
C     XTOL      X tolerance used
C
C     formal parameters
      EXTERNAL FCN
      REAL*8 FCN,TOL(2),XL,XR,X,F
      INTEGER*4 RC
C
C     local variables
      REAL*8 FL,FR,FTOL,XTOL
C
C ------------------------------------------------------------------
C
C     set convergence tolerances
      XTOL=DMAX1(0.D0,TOL(1))
      FTOL=DMAX1(0.D0,TOL(2))
C
C     find the function values at the endpoints
      FL=FCN(XL)
      FR=FCN(XR)
      NF=2
C
C     permit only 100 bisections
      DO 1 I=1,100
C          bisect the interval
           X=0.5D0*(XL+XR)
C
C          evaluate the function at the new point
           F=FCN(X)
           NF=NF+1
C
C          check for convergence in both X and F
           IF(DABS(XR-XL).LE.XTOL .AND. DABS(F).LE.FTOL) THEN
              RC=0
              GO TO 2
           ENDIF
C
C          check for too many function evaluations
           IF(NF.GE.NFE) THEN
              IF(RC.GE.0) WRITE(0,991) NF
  991         FORMAT('Used ',I3,' function evaluations in BISECT')
              RC=3
              GO TO 2
           ENDIF
C
C          decide which side the root is on and update the endpoints
           IF(F*FL .LT. 0.D0) GO TO 3
           IF(F*FR .LT. 0.D0) GO TO 4
C
C          can't continue; check for convergence in X or F
           IF(DABS(XR-XL).LE.XTOL .OR. DABS(F).LE.FTOL) THEN
              RC=1
              GO TO 2
           ENDIF
C
C          the function does not appear to cross zero
           IF(RC.GE.0) WRITE(0,992)
  992      FORMAT('No root on interval in BISECT')
           RC=2
           GO TO 2
C
C          the root is in the left half
    3      XR=X
           FR=F
           GO TO 1
C
C          the root is in the right half
    4      XL=X
           FL=F
    1 CONTINUE
C
C     convergence was not attained in 100 iterations
      IF(RC.GE.0) WRITE(0,993)
  993 FORMAT('No convergence in 100 iterations in BISECT')
      RC=4
    2 NFE=NF
      RETURN
      END
