C
Code by Michael Kupferschmid
C
      SUBROUTINE DFT(NN,INVDIR,SCALE,SHIFT, DATA,B, RC)
C     This routine returns in-place the direct or inverse discrete
C     Fourier transform of the sequence in DATA.
C
C     variable  meaning
C     --------  -------
C     B         upper limit of integral for transform or inverse
C     CDEXP     Fortran function returns exp(COMPLEX*16)
C     DATA      input sequence, then its transform or inverse
C     DCMPLX    Fortran function returns COMPLEX*16 for two REAL*8s
C     DELTA     sampling interval
C     DFLOAT    Fortran function returns REAL*8 for INTEGER*4
C     F         the input f_k's as a COMPLEX*16s
C     FF        a point F_n in the transform
C     I         i, the square root of -1
C     INVDIR    +1 => direct transform, -1 => inverse transform
C     K         index on input values
C     N         index on transform values
C     NN        number of samples
C     PI        the circle constant
C     RC        return code; 0 => parameters make sense
C     SCALE     T => scale output or assume that input is scaled
C     SF        scale factor
C     SHIFT     T => frequency-shift result
C     TEMP      temporary scalar used in shifting result
C     W         exp(i*2*pi/N)
C
C     formal parameters
      LOGICAL*4 SCALE,SHIFT
      COMPLEX*16 DATA(0:NN)
      REAL*8 B
      INTEGER*4 RC
C
C     automatic workspace
      COMPLEX*16 F(0:NN)
C
C     other local variables
      COMPLEX*16 FF,I/(0.D0,1.D0)/,W,TEMP
      REAL*8 DELTA,PI/3.1415926535897932D0/,SF
C
C ------------------------------------------------------------------
C
C     sanity-check the input parameters
      RC=0
C       the number of points must be positive
        IF(NN.LE.1) RC=1
C
C       the transform flag must denote either direct or inverse
        IF(INVDIR.NE.+1 .AND. INVDIR.NE.-1) RC=RC+2
C
C       the upper limit of integration was assumed positive
        IF(B.LE.0.D0) RC=RC+4
      IF(RC.NE.0) RETURN
C
C     save the input sequence
      DO 1 K=0,NN-1
           F(K)=DATA(K)
    1 CONTINUE
C
C     find the complex constant whose powers appear in the series
      W=CDEXP(I*DCMPLX(2.D0*PI/DFLOAT(INVDIR*NN),0.D0))
C
C     compute each F_n from the DFT series
      DO 2 N=0,NN-1
           FF=(0.D0,0.D0)
           DO 3 K=0,NN-1
                FF=FF+W**(N*K)*F(K)
    3      CONTINUE
           DATA(N)=FF
    2 CONTINUE
C
C     find sampling interval and new upper limit of integration
      IF(SCALE) THEN
         IF(INVDIR.EQ.+1) THEN
            DELTA=B/DFLOAT(NN-1)
            IF(     SHIFT) B=DFLOAT(NN/2)/(DELTA*DFLOAT(NN))
            IF(.NOT.SHIFT) B=DFLOAT(NN-1)/(DELTA*DFLOAT(NN))
         ELSE
            IF(     SHIFT) DELTA=2.D0*B/DFLOAT(NN)
            IF(.NOT.SHIFT) DELTA=B/DFLOAT(NN-1)
            B=DFLOAT(NN-1)/(DELTA*DFLOAT(NN))
         ENDIF
         SF=DELTA
      ELSE
         IF(INVDIR.EQ.+1) SF=1.D0
         IF(INVDIR.EQ.-1) SF=1.D0/DFLOAT(NN)
      ENDIF
C
C     scale the output values
      IF(SF.NE.1.D0) THEN
         DO 11 N=0,NN-1
               DATA(N)=SF*DATA(N)
   11    CONTINUE
      ENDIF
C
C     shift, or compensate for having shifted before
      IF(SHIFT) THEN
         IF(INVDIR.EQ.1) THEN
C           frequency-shift the forward transform
            DATA(NN)=DATA(NN/2)
            DO 12 N=0,NN/2-1
                  TEMP=DATA(N)
                  DATA(N)=DATA(N+NN/2)
                  DATA(N+NN/2)=TEMP
   12       CONTINUE
         ELSE
C           fix up the inverse of the frequency-shifted transform
            DO 13 N=1,NN-1,2
                  DATA(N)=-DATA(N)
   13       CONTINUE
         ENDIF
      ENDIF
      RETURN
      END
