C
Code by Michael Kupferschmid
C
      SUBROUTINE DFT2(LX,LY,NNX,NNY,INVDIR,SCALE,SHIFT, F,BX,BY, RC)
C     This routine returns in-place the direct or inverse discrete
C     Fourier transform of the NX x NY array of numbers in the upper
C     left corner of F(LX,LY) by evaluating the nested sum.
C
C     variable  meaning
C     --------  -------
C     BX        limit of integration in the X direction
C     BY        limit of integration in the Y direction
C     CDEXP     Fortran function returns exp(COMPLEX*16)
C     DCMPLX    Fortran function returns COMPLEX*16 for two REAL*8s
C     DFLOAT    Fortran function returns REAL*8 for INTEGER*4
C     DX        sampling interval in Y
C     DY        sampling interval in X
C     F         input data, then its transform or inverse
C     G         intermediate matrix
C     I         i, the square root of -1
C     INVDIR    +1 => direct transform, -1 => inverse transform
C     KX        index on input rows
C     KY        index on input columns
C     LX        first dimension of F in calling routine
C     LY        second dimension of F in calling routine
C     MOD       Fortran function for remainder of INTEGER*4 division
C     NNX       number of rows
C     NNY       number of columns
C     NX        index on output rows
C     NY        index on output columns
C     PI        the circle constant
C     RC        return code; 0 => parameters make sense
C     SCALE     T => scale output or assume that input is scaled
C     SFX       scale factor in X
C     SFY       scale factor in Y
C     SHIFT     T => frequency-shift result
C     TEMP      temporary used in shifting
C     WX        exp(i*2*pi/NNX)
C     WY        exp(i*2*pi/NNY)
C
C     formal parameters
      LOGICAL*4 SCALE,SHIFT
      COMPLEX*16 F(0:LX-1,0:LY-1)
      REAL*8 BX,BY
      INTEGER*4 RC
C
C     automatic workspace
      COMPLEX*16 G(0:LX-1,0:LY-1) 
C
C     other local variables
      COMPLEX*16 I/(0.D0,1.D0)/,TEMP,WX,WY
      REAL*8 DX,DY,PI/3.1415926535897932D0/,SFX,SFY
C
C ------------------------------------------------------------------
C
C     sanity-check the input parameters
      RC=0
C       the numbers of points must be positive
        IF(NNX.LE.1 .OR. NNY.LE.1) RC=1
C
C       the transform flag must denote either direct or inverse
        IF(INVDIR.NE.+1 .AND. INVDIR.NE.-1) RC=RC+2
C
C       the upper limits of integration were assumed positive
        IF(BX.LE.0.D0 .OR. BY.LE.0.D0) RC=RC+4
C
C       the leading dimensions must be big enough
        IF(LX-1.LT.NNX .OR. LY-1.LT.NNY) RC=RC+8
      IF(RC.NE.0) RETURN
C
C     find the complex constant whose powers appear in the series
      WX=CDEXP(I*DCMPLX(2.D0*PI/DFLOAT(INVDIR*NNX),0.D0))
      WY=CDEXP(I*DCMPLX(2.D0*PI/DFLOAT(INVDIR*NNY),0.D0))
C
C     transform the columns
      DO 1 KY=0,NNY-1
      DO 1 NX=0,NNX-1
           G(NX,KY)=(0.D0,0.D0)
           DO 2 KX=0,NNX-1
                G(NX,KY)=G(NX,KY)+F(KX,KY)*WX**(NX*KX)
    2      CONTINUE
    1 CONTINUE
C
C     transform the rows
      DO 3 NX=0,NNX-1
      DO 3 NY=0,NNY-1
           F(NX,NY)=(0.D0,0.D0)
           DO 4 KY=0,NNY-1
                F(NX,NY)=F(NX,NY)+G(NX,KY)*WY**(NY*KY)
    4      CONTINUE
    3 CONTINUE
C
C     find sampling interval and new upper limit of integration
      IF(SCALE) THEN
         IF(INVDIR.EQ.+1) THEN
            DX=BX/DFLOAT(NNX-1)
            DY=BY/DFLOAT(NNY-1)
            IF(SHIFT) THEN
               BX=DFLOAT(NNX/2)/(DX*DFLOAT(NNX))
               BY=DFLOAT(NNY/2)/(DY*DFLOAT(NNY))
            ELSE
               BX=DFLOAT(NNX-1)/(DX*DFLOAT(NNX))
               BY=DFLOAT(NNY-1)/(DY*DFLOAT(NNY))
            ENDIF
         ELSE
            IF(SHIFT) THEN
               DX=2.D0*BX/DFLOAT(NNX)
               DY=2.D0*BY/DFLOAT(NNY)
            ELSE
               DX=BX/DFLOAT(NNX-1)
               DY=BY/DFLOAT(NNY-1)
            ENDIF
            BX=DFLOAT(NNX-1)/(DX*DFLOAT(NNX))
            BY=DFLOAT(NNY-1)/(DY*DFLOAT(NNY))
         ENDIF
         SFX=DX
         SFY=DY
      ELSE
         IF(INVDIR.EQ.+1) THEN
            SFX=1.D0
            SFY=1.D0
         ELSE
            SFX=1.D0/DFLOAT(NNX)
            SFY=1.D0/DFLOAT(NNY)
         ENDIF
      ENDIF
C
C     scale the output values
      IF(SFX.NE.1.D0 .OR. SFY.NE.1.D0) THEN
         DO 5 NY=0,NNY-1
         DO 5 NX=0,NNX-1
              F(NX,NY)=DCMPLX(SFX*SFY,0.D0)*F(NX,NY)
    5    CONTINUE
      ENDIF
C
C     shift, or compensate for having shifted before
      IF(SHIFT) THEN
         IF(INVDIR.EQ.1) THEN
C           frequency-shift forward transforms of columns
            DO 6 NY=0,NNY-1
                 F(NNX,NY)=F(NNX/2,NY)
                 DO 7 NX=0,NNX/2-1
                      TEMP=F(NX,NY)
                      F(NX,NY)=F(NX+NNX/2,NY)
                      F(NX+NNX/2,NY)=TEMP
    7            CONTINUE
    6       CONTINUE
C           frequency-shift forward transforms of rows
            DO 8 NX=0,NNX
                 F(NX,NNY)=F(NX,NNY/2)
                 DO 9 NY=0,NNY/2-1
                      TEMP=F(NX,NY)
                      F(NX,NY)=F(NX,NY+NNY/2)
                      F(NX,NY+NNY/2)=TEMP
    9            CONTINUE
    8       CONTINUE
         ELSE
C           fix up the inverse of the frequency-shifted transform
            DO 10 NY=0,NNY-1
            DO 10 NX=0+MOD(NY+1,2),NNX-1,2
                  F(NX,NY)=-F(NX,NY)
   10       CONTINUE
         ENDIF
      ENDIF
      RETURN
      END
