C
Code by Michael Kupferschmid
C
      SUBROUTINE DR250(N,X)
C     This routine generates a vector X of N normalized
C     double-precision pseudorandom numbers in the interval
C
C     [0,1]
C
C     by using the following recurrence:
C
C        x  = x      xor  x        k = 1...N
C         k    k-147       k-250
C
C     When N is nonpositive on input, the routine fills an internal
C     shift register with 250 starting numbers by using a mixed
C     congruential generator; when N < 0, |N| is used as the seed
C     for this process, and when N = 0 a seed is obtained from the
C     system clock.  On calls with N > 0, new values are returned in
C     X and the generator reseeds itself.
C
C     variable  meaning 
C     --------  -------
C     A         multiplier for mixed congruential algorithm
C     C         increment for mixed congruential algorithm
C     GETIMEOFDAY  unix routine returns time of day and time zone
C     IABS      Fortran function gives |INTEGER*4|
C     IAND      Fortran function gives bitwise AND of fullwords
C     IEOR      Fortran function gives bitwise XOR of fullwords
C     IOR       Fortran function gives bitwise OR of fullwords
C     ISR       internal shift register
C     IX        a doubleword of X as 2 singlewords
C     JX        IX with the words switched
C     K         index on random numbers generated 
C     KM147     index in ISR of doubleword K-147
C     KM250     index in ISR of doubleword K-250
C     MASKC     to set characteristic of ISR values generated 
C     MASKX     to zero characteristic of ISR values generated
C     MASKZ     to zero the high-order bit of a word
C     N         number of random values needed, or seed (see above)
C     NODD      to make SEED end-bits ones
C     TOD       time-of-day [seconds,microseconds] 
C     SEED      seed used for mixed congruential algorithm
C     SET       T => N has been set to -SEED for return
C     X         vector of random values returned
C     XK        a doubleword of X
C     ZONE      unused; for GETIMEOFDAY
C 
C     formal parameter
      REAL*8 X(N)
C
C     prepare to seed the generator
      INTEGER*4 SEED,TOD(2),ZONE(2),NODD/Z'80000001'/
      INTEGER*4 A/843314861/,C/453816693/
      INTEGER*4 MASKZ/Z'7FFFFFFF'/,MASKX/Z'000FFFFF'/
      LOGICAL*4 SET
C
C     prepare to run the generator 
      INTEGER*4 KM250/251/,KM147/148/,MASKC/Z'3FF00000'/
      REAL*8 XK
      INTEGER*4 IX(2),JX(2)
      EQUIVALENCE(XK,IX(1))
C
C     internal shift register with values from SEED=123457 
      INTEGER*4 ISR(500)/
     ;          Z'0007E8AF',Z'D4C00D62',
     ;          Z'000EE2CD',Z'0DDA19B8',
     ;          Z'0008A81B',Z'AEF3F3FE',
     ;          Z'000A5C19',Z'A9113DB4',
     ;          Z'00094E47',Z'5CE7F45A',
     ;          Z'0002DA25',Z'52C3D170',
     ;          Z'00074733',Z'0278AA76',
     ;          Z'000FA8F1',Z'48FBD0EC',
     ;          Z'0006BEDF',Z'7AFC7252',
     ;          Z'000CD47D',Z'EB92F828',
     ;          Z'000EA14B',Z'2BDF67EE',
     ;          Z'000728C9',Z'2C2EC324',
     ;          Z'00049A77',Z'B0FF674A',
     ;          Z'000831D5',Z'2DFC6DE0',
     ;          Z'00091663',Z'FEC80C66',
     ;          Z'00023BA1',Z'912CF45C',
     ;          Z'0003410F',Z'D60EB342',
     ;          Z'000C522D',Z'A6311298',
     ;          Z'000D067B',Z'A5AE77DE',
     ;          Z'000E4179',Z'04B54494',
     ;          Z'000F12A7',Z'35E4363A',
     ;          Z'000C9585',Z'425DC650',
     ;          Z'000CD193',Z'B96A8A56',
     ;          Z'00049A51',Z'454293CC',
     ;          Z'00006F3F',Z'A055D032',
     ;          Z'00085BDD',Z'EE2B6908',
     ;          Z'0006D7AB',Z'48B023CE',
     ;          Z'000AA629',Z'968BC204',
     ;          Z'000BB6D7',Z'E8D5612A',
     ;          Z'000B0535',Z'9E3EDAC0',
     ;          Z'000578C3',Z'CF8F2446',
     ;          Z'0001C501',Z'8503AF3C',
     ;          Z'0001496F',Z'55F0C922',
     ;          Z'000BF18D',Z'4BB8FB78',
     ;          Z'000F14DB',Z'1EF36BBE',
     ;          Z'000756D9',Z'19593B74',
     ;          Z'000D8707',Z'80D1E81A',
     ;          Z'000E80E5',Z'DFB6AB30',
     ;          Z'00060BF3',Z'B424DA36',
     ;          Z'0004BBB1',Z'7BF746AC',
     ;          Z'0008CF9F',Z'A4BE9E12',
     ;          Z'0002133D',Z'0ED0C9E8',
     ;          Z'0008BE0B',Z'80474FAE',
     ;          Z'000F5389',Z'8884B0E4',
     ;          Z'00078337',Z'5E98CB0A',
     ;          Z'00020895',Z'249C37A0',
     ;          Z'00018B23',Z'9FDAAC26',
     ;          Z'00087E61',Z'51645A1C',
     ;          Z'000A01CF',Z'DC5E4F02',
     ;          Z'0005C0ED',Z'BF29D458',
     ;          Z'0006D33B',Z'823ACF9E',
     ;          Z'000D9C39',Z'13352254',
     ;          Z'000CAB67',Z'7CA909FA',
     ;          Z'00009C45',Z'7A868010',
     ;          Z'000AF653',Z'011F9A16',
     ;          Z'00080D11',Z'1851E98C',
     ;          Z'0007DFFF',Z'5E2EDBF2',
     ;          Z'0001FA9D',Z'8C3B1AC8',
     ;          Z'000C546B',Z'E81CEB8E',
     ;          Z'000D30E9',Z'0C518FC4',
     ;          Z'000FFF97',Z'5F41A4EA',
     ;          Z'00053BF5',Z'4ECC8480',
     ;          Z'00054D83',Z'EC22A406',
     ;          Z'000E67C1',Z'BF86F4FC',
     ;          Z'00056A2F',Z'0D4F44E2',
     ;          Z'0001C04D',Z'3D3B9D38',
     ;          Z'000C419B',Z'12FCA37E',
     ;          Z'00091199',Z'5A80F934',
     ;          Z'00047FC7',Z'84619BDA',
     ;          Z'000AE7A5',Z'5E8544F0',
     ;          Z'000390B3',Z'0AD2C9F6',
     ;          Z'00068E71',Z'818A7C6C',
     ;          Z'0005A05F',Z'3E9E89D2',
     ;          Z'000011FD',Z'A1225BA8',
     ;          Z'00099ACB',Z'71A8F76E',
     ;          Z'000C3E49',Z'E82A5EA4',
     ;          Z'000D2BF7',Z'D3C7EECA',
     ;          Z'000C9F55',Z'2687C160',
     ;          Z'0008BFE3',Z'8CDF0BE6',
     ;          Z'000B8121',Z'54A37FDC',
     ;          Z'000B828F',Z'28BBAAC2',
     ;          Z'0007EFAD',Z'FEA65618',
     ;          Z'00075FFB',Z'70B0E75E',
     ;          Z'0001B6F9',Z'1374C014',
     ;          Z'000D0427',Z'8EF39DBA',
     ;          Z'00056305',Z'D36AF9D0',
     ;          Z'0007DB13',Z'97B669D6',
     ;          Z'00083FD1',Z'5AD8FF4C',
     ;          Z'000A10BF',Z'D405A7B2',
     ;          Z'0004595D',Z'043E8C88',
     ;          Z'0008912B',Z'6A63734E',
     ;          Z'00047BA9',Z'1E471D84',
     ;          Z'00070857',Z'C123A8AA',
     ;          Z'000032B5',Z'3185EE40',
     ;          Z'0003E243',Z'3687E3C6',
     ;          Z'0007CA81',Z'51F1FABC',
     ;          Z'00044AEF',Z'8A9B80A2',
     ;          Z'00004F0D',Z'B821FEF8',
     ;          Z'00002E5B',Z'16CF9B3E',
     ;          Z'000F8C59',Z'1E4876F4',
     ;          Z'000E3887',Z'2F570F9A',
     ;          Z'00080E65',Z'1CEF9EB0',
     ;          Z'000FD573',Z'CA4279B6',
     ;          Z'00052131',Z'8375722C',
     ;          Z'000D311F',Z'C85C3592',
     ;          Z'0006D0BD',Z'6847AD68',
     ;          Z'0001378B',Z'FBC45F2E',
     ;          Z'000DE909',Z'6CDFCC64',
     ;          Z'000594B7',Z'484CD28A',
     ;          Z'0007F615',Z'717F0B20',
     ;          Z'000EB4A3',Z'F9952BA6',
     ;          Z'000B43E1',Z'B4AA659C',
     ;          Z'0007C34F',Z'AAE6C682',
     ;          Z'0002DE6D',Z'9A6697D8',
     ;          Z'000EACBB',Z'DCD0BF1E',
     ;          Z'000A91B9',Z'17341DD4',
     ;          Z'00001CE7',Z'9483F17A',
     ;          Z'000AE9C5',Z'7ACB3390',
     ;          Z'00037FD3',Z'20EEF996',
     ;          Z'00053291',Z'9697D50C',
     ;          Z'0007017F',Z'E19A3372',
     ;          Z'000F781D',Z'7BF5BE48',
     ;          Z'000B8DEB',Z'AB43BB0E',
     ;          Z'00008669',Z'CE2C6B44',
     ;          Z'0000D117',Z'A63B6C6A',
     ;          Z'000BE975',Z'E42B1800',
     ;          Z'00013703',Z'427EE386',
     ;          Z'000DED41',Z'B604C07C',
     ;          Z'000DEBAF',Z'1D957C62',
     ;          Z'00079DCD',Z'522C20B8',
     ;          Z'000ADB1B',Z'762C52FE',
     ;          Z'000AC719',Z'566FB4B4',
     ;          Z'000AB147',Z'8972435A',
     ;          Z'0005F525',Z'28B5B870',
     ;          Z'000ADA33',Z'F633E976',
     ;          Z'000073F1',Z'6B7827EC',
     ;          Z'000F81DF',Z'01B7A152',
     ;          Z'00064F7D',Z'6A00BF28',
     ;          Z'000F944B',Z'5A5986EE',
     ;          Z'000453C9',Z'F864FA24',
     ;          Z'0000BD77',Z'B3E7764A',
     ;          Z'00040CD5',Z'034214E0',
     ;          Z'00036963',Z'D9BD0B66',
     ;          Z'0007C6A1',Z'4B390B5C',
     ;          Z'000EC40F',Z'929FA242',
     ;          Z'00068D2D',Z'082A9998',
     ;          Z'000CB97B',Z'725A56DE',
     ;          Z'00082C79',Z'70333B94',
     ;          Z'0005F5A7',Z'751A053A',
     ;          Z'00013085',Z'DE672D50',
     ;          Z'000DE493',Z'00894956',
     ;          Z'000EE551',Z'954E6ACC',
     ;          Z'000EB23F',Z'A6AC7F32',
     ;          Z'000356DD',Z'D920B008',
     ;          Z'00054AAB',Z'467DC2CE',
     ;          Z'00015129',Z'5DC17904',
     ;          Z'000D59D7',Z'E648F02A',
     ;          Z'00086035',Z'C47C01C0',
     ;          Z'000D4BC3',Z'63C7A346',
     ;          Z'0000D001',Z'A57F463C',
     ;          Z'00024C6F',Z'55FD3822',
     ;          Z'0007AC8D',Z'E11A0278',
     ;          Z'000C47DB',Z'BCD2CABE',
     ;          Z'000AC1D9',Z'34B6B274',
     ;          Z'0009EA07',Z'5A73371A',
     ;          Z'00049BE5',Z'CF979230',
     ;          Z'00049EF3',Z'D2671936',
     ;          Z'000886B1',Z'63529DAC',
     ;          Z'000C929F',Z'6A70CD12',
     ;          Z'000E8E3D',Z'EC0D90E8',
     ;          Z'0004B10B',Z'09286EAE',
     ;          Z'000F7E89',Z'AC79E7E4',
     ;          Z'000EA637',Z'CE57DA0A',
     ;          Z'0000E395',Z'9990DEA0',
     ;          Z'0006DE23',Z'E116AB26',
     ;          Z'00010961',Z'320F711C',
     ;          Z'000084CF',Z'4FA63E02',
     ;          Z'0002FBED',Z'7DB25B58',
     ;          Z'0001863B',Z'1D0DAE9E',
     ;          Z'000A8739',Z'30321954',
     ;          Z'000E8E67',Z'D875D8FA',
     ;          Z'00083745',Z'2BFEE710',
     ;          Z'00070953',Z'DA455916',
     ;          Z'00055811',Z'60BCC08C',
     ;          Z'000122FF',Z'82FC8AF2',
     ;          Z'000FF59D',Z'417F61C8',
     ;          Z'0005C76B',Z'97D18A8E',
     ;          Z'0006DBE9',Z'CEC646C4',
     ;          Z'000CA297',Z'990C33EA',
     ;          Z'000596F5',Z'F038AB80',
     ;          Z'00082083',Z'AE222306',
     ;          Z'000072C1',Z'1A218BFC',
     ;          Z'00016D2F',Z'0392B3E2',
     ;          Z'00007B4D',Z'FAABA438',
     ;          Z'0004749B',Z'B683027E',
     ;          Z'000F7C99',Z'AADD7034',
     ;          Z'000BE2C7',Z'2A19EADA',
     ;          Z'000402A5',Z'1F552BF0',
     ;          Z'000D23B3',Z'E29C08F6',
     ;          Z'000D5971',Z'D4C4D36C',
     ;          Z'0004635F',Z'C247B8D2',
     ;          Z'000F8CFD',Z'742E22A8',
     ;          Z'00008DCB',Z'C3F1166E',
     ;          Z'000F6949',Z'EADE95A4',
     ;          Z'000F4EF7',Z'8F5DFDCA',
     ;          Z'000E7A55',Z'322B6860',
     ;          Z'000912E3',Z'03620AE6',
     ;          Z'00070C21',Z'42ED96DC',
     ;          Z'000D058F',Z'91BA99C2',
     ;          Z'00082AAD',Z'F0BDDD18',
     ;          Z'000D12FB',Z'08AAC65E',
     ;          Z'0001A1F9',Z'A8F0B714',
     ;          Z'0009E727',Z'26576CBA',
     ;          Z'000FFE05',Z'D15260D0',
     ;          Z'000EEE13',Z'91E328D6',
     ;          Z'00088AD1',Z'42A2D64C',
     ;          Z'000E53BF',Z'164A56B2',
     ;          Z'0005545D',Z'1AD1D388',
     ;          Z'000D042B',Z'BAFF124E',
     ;          Z'000126A9',Z'E2FAD484',
     ;          Z'000EAB57',Z'164537AA',
     ;          Z'00038DB5',Z'C5211540',
     ;          Z'0001B543',Z'754E62C6',
     ;          Z'000CD581',Z'CDAB91BC',
     ;          Z'000B4DEF',Z'3615EFA2',
     ;          Z'00020A0D',Z'F4A105F8',
     ;          Z'0003615B',Z'EEFCFA3E',
     ;          Z'0008F759',Z'EAA3EDF4',
     ;          Z'00009B87',Z'40265E9A',
     ;          Z'00042965',Z'65AE85B0',
     ;          Z'00046873',Z'EA92B8B6',
     ;          Z'000EEC31',Z'E98EC92C',
     ;          Z'0006F41F',Z'88FC6492',
     ;          Z'00094BBD',Z'C8227468',
     ;          Z'00032A8B',Z'86737E2E',
     ;          Z'00041409',Z'55530364',
     ;          Z'0002B7B7',Z'2EB9E18A',
     ;          Z'000CD115',Z'8AD1B220',
     ;          Z'000A07A3',Z'F45F2AA6',
     ;          Z'0009CEE1',Z'97937C9C',
     ;          Z'0004464F',Z'C89CB582',
     ;          Z'0006196D',Z'970D1ED8',
     ;          Z'000F5FBB',Z'20F19E1E',
     ;          Z'000D7CB9',Z'EC2F14D4',
     ;          Z'0007FFE7',Z'067EC07A',
     ;          Z'000884C5',Z'7C219A90',
     ;          Z'000592D3',Z'4B22B896',
     ;          Z'00087D91',Z'C4C0AC0C',
     ;          Z'0006447F',Z'4055E272',
     ;          Z'0003731D',Z'8AD80548'/
C
C ------------------------------------------------------------------
C
C     is this an initialization call?
      IF(N.LE.0) GO TO 1
C
C     generate N new random numbers while reseeding the generator
      DO 2 K=1,N 
C          find the indices in ISR of X(K-250) and X(K-147)
           KM250=KM250-1
           IF(KM250.LE.0) KM250=250
           KM147=KM147-1
           IF(KM147.LE.0) KM147=250
C
C          exclusive-or the singlewords 
           ISR(2*KM250  )=IEOR(ISR(2*KM250  ),ISR(2*KM147  ))
           ISR(2*KM250-1)=IEOR(ISR(2*KM250-1),ISR(2*KM147-1))
C
C          extract the resulting doubleword
           JX(1)=IOR(ISR(2*KM250-1),MASKC)
           JX(2)=ISR(2*KM250)
C          the word order is reversed in the Pentium
           IX(1)=JX(2)
           IX(2)=JX(1)
           X(K)=XK-1.D0
    2 CONTINUE
      RETURN
C
C     get a seed to use in the reloading process
    1 SET=.FALSE.
      SEED=IABS(N)
      IF(SEED.GT.0) GO TO 3
      CALL GETIMEOFDAY(TOD,ZONE) 
      TOD(1)=A*TOD(1)+C
      TOD(2)=A*TOD(2)+C
      N=IEOR(TOD(1),TOD(2))
      N=IOR(N,NODD)
      SEED=-N 
      SET=.TRUE.
C
C     reload the ISR using a mixed congruential algorithm 
    3 DO 4 K=1,250
C          low-order word of the 8-byte value
           SEED=A*SEED+C
           ISR(2*K)=SEED
           SEED=IAND(SEED,MASKZ)
C
C          high-order word of the 8-byte value
           SEED=A*SEED+C
           ISR(2*K-1)=IAND(SEED,MASKX)
           SEED=IAND(SEED,MASKZ)
    4 CONTINUE
      IF(.NOT.SET) N=-SEED
      KM250=251
      KM147=148
      RETURN
      END
