C
Code by Michael Kupferschmid
C
      SUBROUTINE FFT(N,INVDIR,SCALE,SHIFT, DATA,B, RC)
C     This routine computes in-place the direct or inverse fast
C     Fourier transform of the sequence in DATA.
C
C     variable  meaning
C     --------  -------
C     B         upper limit of integral for transform or inverse
C     DATA      input sequence, then its transform or inverse
C     DCOS      Fortran function returns cosine of REAL*8
C     DELTA     sampling interval or frequency interval
C     DFLOAT    Fortran function returns REAL*8 for INTEGER*4
C     DSIN      Fortran function returns sine of a REAL*8
C     I         index on DATA elements
C     INVDIR    +1 => direct transform, -1 => inverse transform
C     J         second index on DATA elements for rearrangement
C     L         index on powers of W
C     LGN       log_2(N) = number of stages in signal flow graph
C     LMAX      size of the transforms in this stage
C     M         index on stages in the signal flow graph
C     MSZ       bit position of most significant 0 in an index value
C     MZRO      bit pattern 10000000000000000000000000000000
C     N         number of input real-imaginary data pairs
C     NBITS     N bit pattern shifted left
C     NN        number of input data items = 2*N
C     NPREV     previous value of N
C     NS        number of shifts to put N 1 bit in sign position
C     P         odd index of data pair at top of a butterfly
C     PSTEP     DATA elements between butterflies using same W power
C     Q         odd index of data pair at bottom of a butterfly
C     RC        return code; 0 => parameters make sense
C     SCALE     T => scale output or assume that input is scaled
C     SF        scale factor
C     SHIFT     T => frequency-shift result
C     TEMPI     temporary storage for real part of a DATA element
C     TEMPR     temporary storage for imag part of a DATA element
C     THETA     angle in W
C     TWOPI     twice the circle constant
C     WI        imaginary part of W
C     WPI       imaginary part of W for this stage
C     WPR       real part of W for this stage
C     WR        real part of W
C     WTEMP     temporary in finding next power of W
C
C     input parameters
      LOGICAL*4 SCALE,SHIFT
      REAL*8 DATA(*),B
      INTEGER*4 RC
C
C     local variables
      REAL*8 DELTA,SF,TEMPI,TEMPR,THETA,WI,WPI,WPR,WR,WTEMP
      REAL*8 TWOPI/6.2831853071795865D0/
      INTEGER*4 MZRO/Z'80000000'/,NPREV/Z'80000000'/,P,PSTEP,Q
C
C ------------------------------------------------------------------
C
C     sanity-check the input parameters
      RC=0
C       the number of points must be a positive power of 2
        IF(N.NE.NPREV) THEN
C          this is a new N; check it and find its lg
           NBITS=N
           DO 2 NS=1,32
                IF(NBITS.LT.0) THEN
                   IF(NBITS.NE.MZRO) GO TO 3
                   IF(NS.EQ.1 .OR. NS.EQ.32) GO TO 3
                   LGN=32-NS
                   GO TO 4
                ENDIF
                NBITS=2*NBITS
    2      CONTINUE
    3      RC=1
    4      IF(RC.EQ.0) THEN
              NPREV=N
           ELSE
              NPREV=MZRO
           ENDIF
        ENDIF
C
C       the transform flag must denote either direct or inverse
        IF(INVDIR.NE.+1 .AND. INVDIR.NE.-1) RC=RC+2
C
C       the upper limit of integration was assumed positive
        IF(B.LE.0.D0) RC=RC+4
      IF(RC.NE.0) RETURN
      NN=2*N
C
C     arrange the input sequence by ascending bit-reversed index
      J=1
      DO 5 I=3,NN-3,2
           MSZ=N
    7      IF(J.LT.MSZ) GO TO 6
              J=J-MSZ
              MSZ=MSZ/2
           GO TO 7
    6      J=J+MSZ
           IF(J.GT.I) THEN
              TEMPR=DATA(J)
              TEMPI=DATA(J+1)
              DATA(J)=DATA(I)
              DATA(J+1)=DATA(I+1)
              DATA(I)=TEMPR
              DATA(I+1)=TEMPI
           ENDIF
    5 CONTINUE
C
C     use the butterfly algorithm to evaluate the signal flow graph
      LMAX=2
      DO 8 M=1,LGN
C          set W^0=1
           WR=1.D0
           WI=0.D0
C
C          compute the W for this stage
           THETA=TWOPI/DFLOAT(INVDIR*LMAX)
           WPR=DCOS(THETA)
           WPI=DSIN(THETA)
C
C          consider each power of W used in this stage
           PSTEP=2*LMAX
           DO 9 L=1,LMAX,2
C               do the butterflies with this power of W in all groups
                DO 10 P=L,NN,PSTEP
                      Q=P+LMAX
                      TEMPR=WR*DATA(Q)-WI*DATA(Q+1)
                      TEMPI=WR*DATA(Q+1)+WI*DATA(Q)
                      DATA(Q)=DATA(P)-TEMPR
                      DATA(Q+1)=DATA(P+1)-TEMPI
                      DATA(P)=DATA(P)+TEMPR
                      DATA(P+1)=DATA(P+1)+TEMPI
   10           CONTINUE
C
C               find the next power of W as W*WP
                WTEMP=WR
                WR=WR*WPR-WI*WPI
                WI=WI*WPR+WTEMP*WPI
    9      CONTINUE
           LMAX=PSTEP
    8 CONTINUE
C
C     find sampling interval and new upper limit of integration
      IF(SCALE) THEN
         IF(INVDIR.EQ.+1) THEN
            DELTA=B/DFLOAT(N-1)
            IF(     SHIFT) B=DFLOAT(N/2)/(DELTA*DFLOAT(N))
            IF(.NOT.SHIFT) B=DFLOAT(N-1)/(DELTA*DFLOAT(N))
         ELSE
            IF(     SHIFT) DELTA=2.D0*B/DFLOAT(N)
            IF(.NOT.SHIFT) DELTA=B/DFLOAT(N-1)
            B=DFLOAT(N-1)/(DELTA*DFLOAT(N))
         ENDIF
         SF=DELTA
      ELSE
         IF(INVDIR.EQ.+1) SF=1.D0
         IF(INVDIR.EQ.-1) SF=1.D0/DFLOAT(N)
      ENDIF
C
C     scale the output values
      IF(SF.NE.1.D0) THEN
         DO 11 I=1,NN,2
               DATA(I)=SF*DATA(I)
               DATA(I+1)=SF*DATA(I+1)
   11    CONTINUE
      ENDIF
C
C     shift, or compensate for having shifted before
      IF(SHIFT) THEN
         IF(INVDIR.EQ.1) THEN
C           frequency-shift the forward transform
            DATA(NN+1)=DATA(N+1)
            DATA(NN+2)=DATA(N+2)
            DO 12 I=1,N,2
                  TEMPR=DATA(I)
                  TEMPI=DATA(I+1)
                  DATA(I)=DATA(I+N)
                  DATA(I+1)=DATA(I+1+N)
                  DATA(I+N)=TEMPR
                  DATA(I+1+N)=TEMPI
   12       CONTINUE
         ELSE
C           fix up the inverse of the frequency-shifted transform
            DO 13 I=3,NN-1,4
                  DATA(I)=-DATA(I)
                  DATA(I+1)=-DATA(I+1)
   13       CONTINUE
         ENDIF
      ENDIF
      RETURN
      END
