C
Code by Michael Kupferschmid
C
      SUBROUTINE FFT2ST(LX,LY,NX,NY,INVDIR,SCALE,SHIFT,
     ;                  DATA,BX,BY, RC)
C     This routine computes in-place the direct or inverse fast
C     Fourier transform of the numbers in DATA, by using strides
C     to address the appropriate elements.  It assumes that DATA
C     points to an LX x LY COMPLEX*16 array F whose upper-left
C     corner contains the NX x NY matrix to be transformed.
C
C     variable  meaning
C     --------  -------
C     BX        limit of integration in X
C     BY        limit of integration in Y
C     DATA      input array, then its transform or inverse 
C     DCOS      Fortran function returns cosine of REAL*8
C     DCMPLX    Fortran function returns COMPLEX*16 for two REAL*8s
C     DFLOAT    Fortran function returns REAL*8 for INTEGER*4
C     DSIN      Fortran function returns sine of a REAL*8
C     DX        sampling interval in X
C     DY        sampling interval in Y
C     I         index on DATA elements
C     IDIM      index on dimensions
C     IINC      increment on I in data rearrangement
C     INVDIR    +1 => direct transform, -1 => inverse transform
C     IPLIM     limit on I in data rearrangement, P in butterflies
C     IST       starting value of I in data rearrangement
C     J         second index on DATA elements for rearrangement
C     K         index on rows or columns for repeating swaps
C     KINC      increment on K in data rearrangement
C     KLIM      limit on K in data rearrangement
C     L         index on powers of W
C     LGN       log_2(NS(IDIM)) = stages in signal flow graph
C     LMAX      size of the transforms at this stage
C     LX        first dimension of F passed for DATA
C     LY        first dimension of F passed for DATA
C     M         index on stages in the signal flow graph
C     MLT       multiplier in formula for Q, P start and increment
C     MSZ       bit position of most significant 0 in an index value
C     MSZST     starting value of MSZ for this dimension
C     MZRO      bit pattern 10000000000000000000000000000000
C     NBITS     NN(IDIM) bit pattern shifted left
C     NN        [NX,NY]
C     NNPREV    values of NN on the previous call
C     NS        shifts to put NN(IDIM) 1 bit in sign position
C     NX        number of samples in X
C     NY        number of samples in Y
C     P         odd index of data pair at top of a butterfly
C     PSTEP     nodes between butterflies using the same power of W
C     Q         odd index of data pair at bottom of a butterfly
C     RC        return code; 0 => parameters make sense
C     SCALE     T => scale output or assume that input is scaled
C     SFX       scale factor in X
C     SFY       scale factor in Y
C     SHIFT     T => frequency-shift output or assume input shifted
C     T         location in DATA of butterfly element Q
C     TEMPI     temporary storage for real part of a DATA element
C     TEMPR     temporary storage for imag part of a DATA element
C     THETA     angle in W
C     TWOPI     twice circle constant
C     WI        imaginary part of W
C     WMI       imaginary part of W for this stage
C     WMR       real part of W for this stage
C     WR        real part of W
C     WTEMP     temporary in finding next power of W
C
C     formal parameters
      LOGICAL*4 SCALE,SHIFT
      REAL*8 DATA(2*LX*LY),BX,BY
      INTEGER*4 RC
C
C     local variables
      REAL*8 DX,DY,SFX,SFY,TEMPI,TEMPR,THETA,WI,WMI,WMR,WR,WTEMP
      REAL*8 TWOPI/6.2831853071795865D0/
      INTEGER*4 NN(2),NNPREV(2)/2*Z'80000000'/,LGN(2)
      INTEGER*4 MZRO/Z'80000000'/,P,PSTEP,Q,T
C
C ------------------------------------------------------------------
C
C     sanity-check the input parameters
      RC=0
C       each way the number of points must be a positive power of 2
        NN(1)=NX
        NN(2)=NY
        DO 1 IDIM=1,2
             IF(NN(IDIM).NE.NNPREV(IDIM)) THEN
C               this is a new NN; check it and find its log_2
                NBITS=NN(IDIM)
                DO 2 NS=1,32
                     IF(NBITS.LT.0) THEN
                        IF(NBITS.NE.MZRO) GO TO 3
                        IF(NS.EQ.1 .OR. NS.EQ.32) GO TO 3
                        LGN(IDIM)=32-NS
                        GO TO 1
                     ENDIF
                     NBITS=2*NBITS
    2           CONTINUE
    3           RC=1
             ENDIF
    1   CONTINUE
        IF(RC.EQ.0) THEN
C          they were both positive powers of 2; remember them
           NNPREV(1)=NX
           NNPREV(2)=NY
        ELSE
           NNPREV(1)=MZRO
           NNPREV(2)=MZRO
        ENDIF
C
C       the transform flag must denote either direct or inverse
        IF(INVDIR.NE.+1 .AND. INVDIR.NE.-1) RC=RC+2
C
C       the upper limit of integration was assumed positive
        IF(BX.LE.0.D0 .OR. BY.LE.0.D0) RC=RC+4
C
C       the adjustable dimensions must be big enough
        IF(LX.LT.NX .OR. LY.LT.NY) RC=RC+8
      IF(RC.NE.0) RETURN
C
C     step over the two dimensions
      DO 5 IDIM=1,2
C          set the parameters for this dimension
           IF(IDIM.EQ.1) THEN
C             going down the first column
              IST=3
              IPLIM=2*NX-3
              IINC=2
              KLIM=2*LX*NY-2*LX
              KINC=2*LX
              MSZST=NX
              MLT=1
           ELSE
C             going across the first row
              IST=1+2*LX
              IPLIM=2*LX*NY-4*LX+1
              IINC=2*LX
              KLIM=2*NX-2
              KINC=2
              MSZST=LX*NY
              MLT=LX
           ENDIF
C
C          arrange input sequence by ascending bit-reversed index
           J=1
           DO 6 I=IST,IPLIM,IINC
                MSZ=MSZST
    8           IF(J.LT.MSZ) GO TO 7
                   J=J-MSZ
                   MSZ=MSZ/2
                GO TO 8
    7           J=J+MSZ
                IF(J.GT.I) THEN
C                  repeat row swaps for all columns
                   DO 9 K=I,I+KLIM,KINC
C                       repeat the pair of row swaps in each column
                        T=J+(K-I)
                        TEMPR=DATA(K)
                        TEMPI=DATA(K+1)
                        DATA(K)=DATA(T)
                        DATA(K+1)=DATA(T+1)
                        DATA(T)=TEMPR
                        DATA(T+1)=TEMPI
    9              CONTINUE
                ENDIF
    6      CONTINUE
C
C          use the butterfly algorithm to evaluate signal flow graph
           LMAX=2
           DO 10 M=1,LGN(IDIM)
C                set W^0=1
                 WR=1.D0
                 WI=0.D0
C
C                compute the W for this stage
                 THETA=TWOPI/DFLOAT(INVDIR*LMAX)
                 WMR=DCOS(THETA)
                 WMI=DSIN(THETA)
C
C                consider each power of W used in this stage
                 PSTEP=2*LMAX*MLT
                 DO 11 L=1,LMAX,2
C                      butterflies for this power of W in all groups
                       DO 12 P=(L-1)*MLT+1,IPLIM,PSTEP
                             Q=P+LMAX*MLT
C                            repeat for all rows or columns
                             DO 13 K=P,P+KLIM,KINC
                                   T=Q+(K-P)
                                   TEMPR=WR*DATA(T)-WI*DATA(T+1)
                                   TEMPI=WR*DATA(T+1)+WI*DATA(T)
                                   DATA(T)=DATA(K)-TEMPR
                                   DATA(T+1)=DATA(K+1)-TEMPI
                                   DATA(K)=DATA(K)+TEMPR
                                   DATA(K+1)=DATA(K+1)+TEMPI
   13                        CONTINUE
   12                  CONTINUE
C
C                      find the next power of W as W*WP
                       WTEMP=WR
                       WR=WR*WMR-WI*WMI
                       WI=WI*WMR+WTEMP*WMI
   11            CONTINUE
                 LMAX=LMAX*2
   10      CONTINUE
    5 CONTINUE
C
C     find sampling intervals and new upper limits of integration
      IF(SCALE) THEN
         IF(INVDIR.EQ.+1) THEN
            DX=BX/DFLOAT(NX-1)
            DY=BY/DFLOAT(NY-1)
            IF(SHIFT) THEN
               BX=DFLOAT(NX/2)/(DX*DFLOAT(NX))
               BY=DFLOAT(NY/2)/(DY*DFLOAT(NY))
            ELSE
               BX=DFLOAT(NX-1)/(DX*DFLOAT(NX))
               BY=DFLOAT(NY-1)/(DY*DFLOAT(NY))
            ENDIF
         ELSE
            IF(SHIFT) THEN
               DX=2.D0*BX/DFLOAT(NX)
               DY=2.D0*BY/DFLOAT(NY)
            ELSE
               DX=BX/DFLOAT(NX-1)
               DY=BY/DFLOAT(NY-1)
            ENDIF
            BX=DFLOAT(NX-1)/(DX*DFLOAT(NX))
            BY=DFLOAT(NY-1)/(DY*DFLOAT(NY))
         ENDIF
         SFX=DX
         SFY=DY
      ELSE
         IF(INVDIR.EQ.+1) THEN
            SFX=1.D0
            SFY=1.D0
         ELSE
            SFX=1.D0/DFLOAT(NX)
            SFY=1.D0/DFLOAT(NY)
         ENDIF
      ENDIF
C
C     scale the output values
      IF(SFX.NE.1.D0 .OR. SFY.NE.1.D0) THEN
         DO 14 I=1,2*LX*LY
               DATA(I)=DCMPLX(SFX*SFY,0.D0)*DATA(I)
   14    CONTINUE
      ENDIF
C
C     shift, or compensate for having shifted before
      IF(SHIFT) THEN
         IF(INVDIR.EQ.1) THEN
C           frequency-shift forward transforms of columns
            DO 15 I=1,2*LX*NY-2*LX+1,2*LX
                  DATA(I+2*NX  )=DATA(I+NX  )
                  DATA(I+2*NX+1)=DATA(I+NX+1)
                  DO 16 J=I,I+NX-1,2
                        TEMPR=DATA(J)
                        TEMPI=DATA(J+1)
                        DATA(J)=DATA(J+NX)
                        DATA(J+1)=DATA(J+1+NX)
                        DATA(J+NX)=TEMPR
                        DATA(J+1+NX)=TEMPI
   16             CONTINUE
   15       CONTINUE
C           frequency-shift forward transforms of rows
            DO 17 I=1,2*(NX+1)-1,2
                  DATA(2*LX*NY+I  )=DATA(I+LX*NY  )
                  DATA(2*LX*NY+I+1)=DATA(I+LX*NY+1)
                  DO 18 J=I,I+LX*(NY-1),2*LX
                        TEMPR=DATA(J)
                        TEMPI=DATA(J+1)
                        DATA(J)=DATA(J+LX*NY)
                        DATA(J+1)=DATA(J+LX*NY+1)
                        DATA(J+LX*NY)=TEMPR
                        DATA(J+LX*NY+1)=TEMPI
   18             CONTINUE
   17       CONTINUE
         ELSE
C           fix up the inverse of the frequency-shifted transform
            DO 19 I=2*LX+1,2*LX*NY-NY+1,4*LX
                  DO 20 J=I,I+2*NX-2,2
                        DATA(J)=-DATA(J)
                        DATA(J+1)=-DATA(J+1)
   20             CONTINUE
   19       CONTINUE
            DO 21 I=3,2*NX-1,4
                  DO 22 J=I,I+2*LX*(NY-1),2*LX
                        DATA(J)=-DATA(J)
                        DATA(J+1)=-DATA(J+1)
   22             CONTINUE
   21       CONTINUE
         ENDIF
      ENDIF
      RETURN
      END
