C
Code by Michael Kupferschmid
C
      SUBROUTINE FFT2TR(LX,LY,NX,NY,INVDIR,SCALE,SHIFT, F,BX,BY, RC)
C     This routine computes in-place the direct or inverse fast
C     Fourier transform of the NX x NY array of numbers in the upper
C     left corner of F(LX,LY) by using calls to FFT and in-place
C     transpositions.
C
C     variable  meaning
C     --------  -------
C     BX        limit of integration in the X direction
C     BY        limit of integration in the Y direction
C     DCMPLX    Fortran function returns COMPLEX*16 for two REAL*8s
C     DFLOAT    Fortran function returns REAL*8 for INTEGER*4
C     DX        sampling interval in X
C     DY        sampling interval in Y
C     F         data matrix
C     FFT       routine computes 1-dimensional FFT
C     I         index on rows of F
C     INVDIR    1 => direct, -1 => inverse
C     J         index on columns of F
C     LX        first dimension of F in calling routine
C     LY        second dimension of F in calling routine
C     MAX0      Fortran function returns larger of INTEGER*4s
C     MOD       Fortran function for remainder of INTEGER*4 division
C     N         size of transposition
C     NX        number of rows
C     NY        number of columns
C     RC        return code; 0 => all went well
C     RCFFT     FFT return code; 0 => all went well
C     SCALE     T => scale transform/transform was scaled
C     SFX       scale factor in X
C     SFY       scale factor in Y
C     SHIFT     T => frequency-shift transform/transform was shifted
C     TEMP      temporary for in-place transpositions
C
C     formal parameters
      LOGICAL*4 SCALE,SHIFT
      COMPLEX*16 F(LX,LY)
      REAL*8 BX,BY
      INTEGER*4 RC
C
C     local variables
      REAL*8 DX,DY,SFX,SFY
      INTEGER*4 RCFFT
      COMPLEX*16 TEMP
C
C ------------------------------------------------------------------
C
C     check that there is enough space in F for the transposes
      N=MAX0(NX,NY)
      IF(NX.LE.0 .OR. NY.LE.0 .OR. N.GT.LX .OR. N.GT.LY) THEN
         RC=8
         RETURN
      ENDIF
C
C     check that there is enough space in F for the result
      IF(SHIFT .AND. (NX+1.GT.LX .OR. NY+1.GT.LY)) THEN
         RC=8
         RETURN
      ENDIF
C
C     transform the columns
      DO 1 J=1,NY
           CALL FFT(NX,INVDIR,.FALSE.,.FALSE., F(1,J),BX, RCFFT)
           IF(RCFFT.NE.0) THEN
              RC=RCFFT
              RETURN
           ENDIF
    1 CONTINUE
C
C     transpose in-place
      DO 2 I=1,N
           DO 3 J=1,I-1
                TEMP=F(I,J)
                F(I,J)=F(J,I)
                F(J,I)=TEMP
    3      CONTINUE
    2 CONTINUE
C
C     transform the rows
      DO 4 I=1,NX
           CALL FFT(NY,INVDIR,.FALSE.,.FALSE., F(1,I),BY, RCFFT)
           IF(RCFFT.NE.0) THEN
              RC=RCFFT
              RETURN
           ENDIF
    4 CONTINUE
C
C     transpose in-place
      DO 5 I=1,N
           DO 6 J=1,I-1
                TEMP=F(I,J)
                F(I,J)=F(J,I)
                F(J,I)=TEMP
    6      CONTINUE
    5 CONTINUE
C
C     find sampling intervals and new upper limits of integration
      IF(SCALE) THEN
         IF(INVDIR.EQ.+1) THEN
            DX=BX/DFLOAT(NX-1)
            DY=BY/DFLOAT(NY-1)
            IF(SHIFT) THEN
               BX=DFLOAT(NX/2)/(DX*DFLOAT(NX))
               BY=DFLOAT(NY/2)/(DY*DFLOAT(NY))
            ELSE
               BX=DFLOAT(NX-1)/(DX*DFLOAT(NX))
               BY=DFLOAT(NY-1)/(DY*DFLOAT(NY))
            ENDIF
            SFX=DX
            SFY=DY
         ELSE
            IF(SHIFT) THEN
               DX=2.D0*BX/DFLOAT(NX)
               DY=2.D0*BY/DFLOAT(NY)
            ELSE
               DX=BX/DFLOAT(NX-1)
               DY=BY/DFLOAT(NY-1)
            ENDIF
            BX=DFLOAT(NX-1)/(DX*DFLOAT(NX))
            BY=DFLOAT(NY-1)/(DY*DFLOAT(NY))
C           scaling for inverse is done inside FFT
            SFX=DX*DFLOAT(NX)
            SFY=DY*DFLOAT(NY)
         ENDIF
      ELSE
C        forward transform is not to be scaled
C        scaling for inverse is done inside FFT
         SFX=1.D0
         SFY=1.D0
      ENDIF
C
C     scale result to match Fourier transform if desired
      IF(SFX.NE.1.D0 .OR. SFY.NE.1.D0) THEN
         DO 7 I=1,NX
         DO 7 J=1,NY
              F(I,J)=DCMPLX(SFX*SFY,0.D0)*F(I,J)
    7    CONTINUE
      ENDIF
C
C     shift, or compensate for having shifted before
      IF(SHIFT) THEN
         IF(INVDIR.EQ.1) THEN
C           frequency-shift forward transforms of columns
            DO 8 J=1,NY
                 F(NX+1,J)=F(1+NX/2,J)
                 DO 9 I=1,NX/2
                      TEMP=F(I,J)
                      F(I,J)=F(I+NX/2,J)
                      F(I+NX/2,J)=TEMP
    9            CONTINUE
    8       CONTINUE
C           frequency-shift forward transforms of rows
            DO 10 I=1,NX+1
                  F(I,NY+1)=F(I,1+NY/2)
                  DO 11 J=1,NY/2
                        TEMP=F(I,J)
                        F(I,J)=F(I,J+NY/2)
                        F(I,J+NY/2)=TEMP
   11             CONTINUE
   10       CONTINUE
         ELSE
C           fix up the inverse of the frequency-shifted transform
            DO 12 J=1,NY
            DO 12 I=1+MOD(J,2),NX,2
                  F(I,J)=-F(I,J)
   12       CONTINUE
         ENDIF
      ENDIF
      RETURN
      END
