C
Code by Michael Kupferschmid
C
      SUBROUTINE HIST(X,N,LPDF, NC, PDF,CDFI,RC)
C     This routine computes the sample probability density function
C     PDF (histogram) and sample inverse cumulative distribution
C     function CDFI of the N observations in X. If NC <= 0 on input,
C     NC is changed on return to the number of cells actually used,
C     as determined by Sturges' rule (below); if NC = 0 on entry,
C     empty cells are not allowed.
C
C     observations  cells               RC  condition
C     ------------  -----               --  ------------------------
C          1          1                  0  all went well
C          2          2                  1  bogus input parameters
C         3-4         3                  2  cell width zero
C         5-8         4
C         9-16        5
C        17-32        6
C          :          :
C          n         1+ceil(lg(n))
C
C     variable  meaning
C     --------  -------
C     A         1/ln(2) from Maple at Digits:=50;
C     CDFI      inverse cumulative distribution function
C     DFLOAT    Fortran function returns REAL*8 for INTEGER*4
C     DLOG      Fortran function returns natural log of a REAL*8
C     I         index on cells
C     ICEIL     function returns ceiling of a REAL*8
C     IFLOR     function returns floor of a REAL*8
C     J         index on data values
C     LPDF      dimensioned size of PDF and CDFI
C     N         number of data values
C     NC        number of cells used
C     NOEMTY    T => empty cells are not allowed
C     PDF       probability density function
C     RC        return code; see table above
C     W         width of each histogram cell
C     X         vector of input data
C     XMAX      highest value in X
C     XMIN      lowest value in X
C
C     formal parameters
      REAL*8 X(N),PDF(LPDF,2),CDFI(LPDF,2)
      INTEGER*4 RC
C
C     local variables
      REAL*8 W,XMAX,XMIN,A/1.4426950408889634D+00/
      LOGICAL*4 NOEMTY
C
C ------------------------------------------------------------------
C
C     do the input parameters make sense?
      IF(N.LT.2) GO TO 1
      IF(LPDF.LT.3) GO TO 1
C
C     if the number of cells is not specified, use Sturges' rule
      IF(NC.LE.0) NC=1+ICEIL(A*DLOG(DFLOAT(N)))
      NOEMTY=(NC.EQ.0)
C
C     is there room for the results?
      IF(LPDF.LT.NC+1) GO TO 1
C
C     find the maximim and minimum data values
      XMIN=X(1)
      XMAX=X(1)
      DO 2 J=1,N
           IF(X(J).GT.XMAX) XMAX=X(J)
           IF(X(J).LT.XMIN) XMIN=X(J)
    2 CONTINUE
C
C     find the cell width and boundary coordinates
    7 W=(XMAX-XMIN)/DFLOAT(NC)
      IF(W.EQ.0.D0) GO TO 3
      DO 4 I=1,NC+1
           PDF(I,1)=XMIN+W*DFLOAT(I-1)
           PDF(I,2)=0.D0
           CDFI(I,2)=PDF(I,1)
    4 CONTINUE
C
C     accumulate the cell frequencies
      DO 5 J=1,N
C          boundary values go in the cell on the right
           I=1+IFLOR((X(J)-XMIN)/W)
C
C          except for the rightmost boundary value
           IF(I.GT.NC) I=NC
C
C          increment the count for this cell
           PDF(I,2)=PDF(I,2)+1.D0
    5 CONTINUE
      PDF(NC+1,2)=0.D0
C
C     remove empty cells if that is required
      IF(NOEMTY) THEN
         DO 6 I=1,NC
              IF(PDF(I,2).EQ.0.D0) THEN
                 NC=NC-1
                 IF(NC.EQ.0) RETURN
                 GO TO 7
              ENDIF
    6    CONTINUE
      ENDIF
C
C     integrate the cell frequencies
      CDFI(1,1)=0.D0
      DO 8 I=1,NC
           CDFI(I+1,1)=CDFI(I,1)+PDF(I,2)*W
    8 CONTINUE
C
C     normalize so CDFI(NC+1,1)=1
      DO 9 I=1,NC
           PDF(I,2)=PDF(I,2)/CDFI(NC+1,1)
           CDFI(I,1)=CDFI(I,1)/CDFI(NC+1,1)
    9 CONTINUE
      CDFI(NC+1,1)=1.D0
      RC=0
      RETURN
C
C     an input parameter did not make sense
    1 RC=1
      RETURN
C
C     the cell width came out zero
    3 RC=2
      RETURN
      END
