C
Code by Michael Kupferschmid
C
      SUBROUTINE L1REGW(X,Y,W,M, BETA,RC)
C     This routine does weighted simple linear least-absolute-value
C     regression to find BETA(1) and BETA(2) in
C
C     y(x)=BETA(1)+BETA(2)*x
C
C     so as to minimize the sum of the weighted absolute deviations
C     |W(I)|*|y(X(I))-Y(I)|  i=1...m
C     by solving the equivalent linear program having this tableau.
C
C              B1P B2P     R     U(1) V(1) U(2) V(2) ... U(M) V(M)
C     +------+-----------------------------------------------------+
C     |   0  |  0   0      0     W(1) W(1) W(2) W(2) ... W(M) W(M) |
C     +------+-----------------------------------------------------+
C     | Y(1) |  1 X(1) -(1+X(1))  1    -1                          |
C     | Y(2) |  1 X(2) -(1+X(2))            1   -1   .             |
C     :      :  :  :      :                            .           :
C     | Y(M) |  1 X(M) -(1+X(M))                          1   -1   |
C     +------+-----------------------------------------------------+
C
C     Then BETA(1)=B1P-R and BETA(2)=B2P-R.
C
C     variable  quantity
C     --------  --------
C     BASIS     row indices of 1s in basic columns, 0 for nonbasic
C     BETA      regression coefficient vector (intercept,slope)
C     BRUV      linear program solution vector
C     DABS      Fortran function returns |REAL*8|
C     DEBUG     T => print the starting tableau
C     DW        the absolute value of a weight
C     I         index on rows of T
C     J         index on columns of T
C     K         index on the elements of W
C     LDT       leading dimension of T
C     M         number of data points = number of constraints
C     MP        M+1 = number of rows in T
C     MR        number of T rows in use (all of them)
C     N         number of variables in the linear program
C     RC        return code from SIMPLX; 0 => all went well
C     ROWS      indices of the rows in use
C     SIMPLX    routine solves a linear program by pivoting
C     SW        the sum of the absolute values of the weights
C     T         linear program tableau
C     W         vector of weights
C     X         vector of independent variable values
C     Y         vector of dependent variable observations
C
C     formal parameters
      REAL*8 X(M),Y(M),W(M),BETA(2)
      INTEGER*4 RC
C
C     automatic arrays to use in solving the linear program
      REAL*8 T(M+1,4+2*M),BRUV(3+2*M)
      INTEGER*4 ROWS(M+1),BASIS(3+2*M)
C
C     other local variables
      LOGICAL*4 DEBUG/.FALSE./
      REAL*8 DW,SW
C
C ------------------------------------------------------------------
C
C     are there any data points?
      BETA(1)=0.
      BETA(2)=0.
      IF(M.LE.0) THEN
         RC=-1
         RETURN
      ENDIF
C
C     is there only one data point?
      IF(M.EQ.1) THEN
         BETA(1)=Y(1)
         RC=0
         RETURN
      ENDIF
C
C     more than one data point; figure out the dimensions of the LP
      N=3+2*M
      MP=M+1
      MR=MP
      DO 1 I=1,MP
           ROWS(I)=I
    1 CONTINUE
C
C     construct the simplex tableau for the LP
C     objective row
        T(1,1)=0.D0
        T(1,2)=0.D0
        T(1,3)=0.D0
        T(1,4)=0.D0
        SW=0.D0
        DO 2 J=1,2*M-1,2
             K=1+J/2
             DW=DABS(W(K))
             T(1,4+J)=DW
             T(1,5+J)=DW
             SW=SW+DW
    2   CONTINUE
C     constraint rows
        DO 3 I=2,MP
             T(I,1)=Y(I-1)
             T(I,2)=1.D0
             T(I,3)=X(I-1)
             T(I,4)=-1.D0-X(I-1)
             DO 4 J=1,2*M-1,2
                  K=1+J/2
                  IF(I.EQ.K+1) THEN
                     T(I,4+J)= 1.D0
                     T(I,5+J)=-1.D0
                  ELSE
                     T(I,4+J)=0.D0
                     T(I,5+J)=0.D0
                  ENDIF
    4        CONTINUE
    3   CONTINUE
C
C     write out the starting tableau
      IF(DEBUG) THEN
         DO 5 I=1,MP
              WRITE(6,901) (T(I,J),J=1,N+1)
  901         FORMAT(100(1X,F5.1))
    5    CONTINUE
      ENDIF
C
C     are there any nonzero weights?
      IF(SW.EQ.0.D0) THEN
         RC=-1
         RETURN
      ENDIF
C
C     solve the LP to compute the least-absolute-values fit
      LDT=M+1
      CALL SIMPLX(T,LDT,MP,N,ROWS,MR, BASIS,BRUV,RC)
C
C     return the regression coefficients
      IF(RC.EQ.0) THEN
         BETA(1)=BRUV(1)-BRUV(3)
         BETA(2)=BRUV(2)-BRUV(3)
      ENDIF
      RETURN
      END
