C
Code by Michael Kupferschmid
C
      SUBROUTINE TIMER(TCC,STA)
C     This routine measures the CPU time used in executing
C     different parts of a program.
C
C     TCC  action taken
C     ---  ------------
C     -1   initialize; reset all bin counts and times to zero
C      0   update counts and times without changing bin number
C     >0   switch timing to bin number TCC
C
C     The parameter STA is used to identify the origin of the
C     TIMER call in the event that TCC has an illegal value
C     or some other error occurs.
C
C     variable  meaning
C     --------  -------
C     BINCPU    vector of visible [sec,nsec] CPU bin counts
C     BINCYC    vector of processor cycle bin counts
C     CPS       number of processor cycles per second
C     DELTA     a cycle count interval
C     DFLOAT    Fortran function gives REAL*8 for INTEGER*4
C     GETCYC    Pentium chip cycle counter, precise to 1 cycle
C     GETMHZ    routine gets processor speed
C     IFIX      Fortran function gives INTEGER*4 for REAL*4
C     INITMX    parameter sets MAXBIN
C     INITOH    initial value for TOH in cycles
C     L         index on the bins
C     MAXBIN    highest permissible bin number
C     NBIN      number of bin just receiving time
C     NCYC      number of cycles in this bin
C     NONALG    number of bin for non-algorithm time
C     NSEC      a number of nanoseconds
C     NSPCYC    nanoseconds per cycle
C     NTMEAS    number of timing measurements made so far
C     OLDTCC    previous value of TCC
C     SECCYC    number of cycles in SECS seconds, as INTEGER*8 
C     SECS      a number of seconds
C     SNGL      Fortran function gives REAL*4 for REAL*8
C     SPEED     processor speed in MHz millions of cycles per second
C     STA       number designating CALL statement in source
C     TCC       timer control code; see table above
C     TE        cycle count at entry to this routine
C     TIMING    T => timing is enabled
C     TL        cycle count at previous entry
C     TOH       overhead correction in cycles
C     TOPBIN    highest bin number used so far
C     TX        cycle count at exit from this routine
C
C     formal parameters
      INTEGER*4 TCC,STA
C
C     communicate timing information in common
      PARAMETER(INITMX=22)
      COMMON /EXPT/ TIMING,NONALG,NBIN,TOPBIN,NTMEAS,TOH,BINCPU,
     ;              SPEED
      REAL*8 SPEED
      INTEGER*4 TOPBIN,TOH,BINCPU(2,INITMX)
      LOGICAL*4 TIMING
C
C     other local variables
      REAL*8 CPS,NSPCYC
      INTEGER*8 TE/0/,TL,TX/0/,DELTA
      INTEGER*8 BINCYC(INITMX),NCYC,SECCYC
      INTEGER*4 OLDTCC/-2/,INITOH/65/,SECS,NSEC
C
C ------------------------------------------------------------------
C
C     do nothing unless timing is enabled
      IF(.NOT.TIMING) RETURN
C
C     get cycle count on entry so calculations here timed separately
      TL=TE
      CALL GETCYC(TE)
C
C     make sure that the first call is to initialize
      IF(OLDTCC.GE.-1) GO TO 1
      IF(TCC.EQ.-1) GO TO 2
      GO TO 3
C
C     this is not the first call; is this an update?
    1 IF(TCC.NE.0) GO TO 4
      IF(OLDTCC.NE.-1) GO TO 5
      GO TO 3
C
C     this is not an update call; if TCC repeats assume it is wrong
    4 IF(TCC.EQ.OLDTCC) GO TO 3
C
C     here TCC should be in the range -1,1...MAXBIN
      IF(TCC.GT.MAXBIN) GO TO 3
      IF(TCC.GT.-1) GO TO 6
      IF(TCC.EQ.-1) GO TO 2
      GO TO 3
C
C     keep track of the highest bin index used so far
    6 IF(OLDTCC.GE.NONALG) GO TO 5
      IF(OLDTCC.GT.TOPBIN) TOPBIN=OLDTCC
      GO TO 5
C
C     reset times (TCC=-1)
    2 MAXBIN=INITMX
      NONALG=MAXBIN-1
      NBIN=NONALG
      TOPBIN=0
      NTMEAS=0
      TOH=INITOH
      CALL GETMHZ(SPEED)
      WRITE(0,901) SPEED
  901 FORMAT('TIMER assumes processor speed is ',F8.3,' MHz.')
      CPS=SPEED*1.D+06
      NSPCYC=1000.D0/SPEED
      DO 7 L=1,MAXBIN
           BINCYC(L)=0
           BINCPU(1,L)=0
           BINCPU(2,L)=0
    7 CONTINUE
      OLDTCC=-1
      GO TO 8
C
C     accumulate processor cycles for bin NBIN (TCC=0,1...MAXBIN)
    5 DELTA=TE-TX-TOH
      IF(DELTA.LT.0) DELTA=0
      BINCYC(NBIN)=BINCYC(NBIN)+DELTA
      NTMEAS=NTMEAS+1
C
C     accumulate processor cycles used making timing measurements
      DELTA=TX-TL+TOH
      BINCYC(MAXBIN)=BINCYC(MAXBIN)+DELTA
C
C     convert the processor cycle counts into CPU times
      DO 9 L=1,MAXBIN
           NCYC=BINCYC(L)
           SECS=IFIX(SNGL(DFLOAT(NCYC)/CPS))
           BINCPU(1,L)=SECS
C
C          explicitly fixing this product can overflow a fullword
C          so convert to long integer implicitly by assignment 
           SECCYC=CPS*DFLOAT(SECS)
C
           NCYC=NCYC-SECCYC
           NSEC=IFIX(0.5+SNGL(NSPCYC*DFLOAT(NCYC)))
           BINCPU(2,L)=NSEC
    9 CONTINUE
C
C     if this is an update leave the bin number alone; else reset it
      IF(TCC.EQ.0) GO TO 8
      NBIN=TCC
      OLDTCC=TCC
C
C     get cycle count upon returning to the instrumented code
    8 CALL GETCYC(TX)
      RETURN
C
C     handle errors
    3 IF(STA.EQ.0) WRITE(0,903)
  903 FORMAT('A call to TIMER')
      IF(STA.NE.0) WRITE(0,904) STA
  904 FORMAT('A call to TIMER at station',I10)
      WRITE(0,905) TCC,OLDTCC
  905 FORMAT('gives the illegal control code',I10/
     ;       'the previous control code was ',I10/)
      STOP
      END
